from __future__ import absolute_import, unicode_literals
from distll_celery import Celery

app = Celery('tasks', broker='amqp://')


@app.task()
def add(x, y):
    return x + y

if __name__ == '__main__':
    app.start()
